/**************************************************************************
 * 
 * PROJECT: Task Control Management (TCM)
 *
 * (c) Copyright 1997 Reid Simmons.  All rights reserved.
 *
 * FILE: tcmResponder.cc
 *
 * ABSTRACT: Functions for responding to TCM invocations from ControlShell
 *
 * EXPORTS:
 *
 * $Revision: 1.3 $
 * $Date: 1998/03/18 14:41:53 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 * $Log: tcmResponder.cc,v $
 * Revision 1.3  1998/03/18 14:41:53  reids
 * Changes to support both SunOS and Solaris versions.
 *
// Revision 1.2  98/01/02  00:21:15  reids
// Initial attempt to integrate TCM and ControlShell via NDDS.
//   A bit kludgy, since I don't really understand how to handle a series
//   of productions -- really want them to be queued, reliably...
// 
// Revision 1.1  97/11/21  14:05:08  reids
// First release of TCM -- seems to be a stable version
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "cs.h"
#include "CSFsm.h"
#include "CSFsmResponseModule.pkg.h"
#include "tcmResponder.h"

/**************************************************************************
 *
 * CLASS: ResponseTransformClass
 *
 **************************************************************************/

ResponseTableClass::ResponseTableClass (int maxElements)
{
  _numElements = 0;
  _maxElements = maxElements;
  if (maxElements > 0)
    _table = new TCM_CS_Response_Type[maxElements];
  else
    _table = NULL;
}

void ResponseTableClass::add (String context, int returnCode,
			      String responseString)
{
  if (_numElements == _maxElements) {
    fprintf(stderr, "ERROR: response table too full for <%s, %d, %s>\n",
	    context, returnCode, responseString);
    exit(-1);
  } else {
    _table[_numElements].set(context, returnCode, responseString);
    _numElements++;
  }
}

String ResponseTableClass::getResponse (String context, int returnCode) const
{
  for (int i=0; i<_numElements; i++) {
    if (returnCode == _table[i].returnCode &&
	streq(context, _table[i].context)) {
      return _table[i].responseString;
    }
  }
  return NULL;
}

/**************************************************************************
 *
 * CLASS: tcmResponderClass
 *
 **************************************************************************/

boolean tcmResponderClass::respond (String context, int returnCode) const
{
  String responseString; 

  responseString = _responseTable.getResponse((String)context, returnCode);
  if (responseString) {
    if (streq(responseString, SUCCESS_RESPONSE)) {
      fprintf(stderr, "tcmResponderClass:respond: %s => SUCCESS\n", context);
      TCM_Success(_treeRef);
    } else {
      fprintf(stderr, "tcmResponderClass:respond: %s => %s\n", 
	      context, responseString);
      TCM_Failure(_treeRef, responseString, NULL);
    }
  }
  return (responseString != NULL);
}

// Respond in case the current stimulus was not handled by the current state.

boolean tcmResponderClass::missingStimulus (String stimulus) const
{
  boolean sameResponder;

  sameResponder = !strcmp(stimulus, _stimulus);
  if (sameResponder) {
    TCM_Failure(_treeRef, NO_STIM_MODULE, NULL);
  }
  return sameResponder;
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type tcmSendStimulus (TCM_Task_Tree_Ref parentNode,
 *				              CSFsm fsm, String stimulus,
 *				              int numResponses,
 *				              TCM_CS_Response_Type *responses)
 *
 * DESCRIPTION: General function for sending a stimulus to CS and setting
 *              up the intention to await response(s) to that stimulus
 *              (all within a single process).
 *
 ****************************************************************/
TCM_Return_Type tcmSendStimulus (TCM_Task_Tree_Ref parentNode, CSFsm fsm,
				 String stimulus, int numResponses,
				 TCM_CS_Response_Type *responses)
{
  tcmResponder responder = new tcmResponderClass(stimulus, parentNode, 
						 numResponses);
  responder->add(fsm);
  for (int i=0; i < numResponses; i++) {
    responder->addResponse(responses[i].context,
			   responses[i].returnCode,
			   responses[i].responseString);
  }

  // Send stimulus
  fsm->StimulusSend(stimulus, 0);

  return TCM_Ok;
}

/**************************************************************************
 *
 * FUNCTION: TCM_Return_Type tcmSendStimulus (TCM_Task_Tree_Ref parentNode,
 *				              CSFsm fsm,
 *				              String stimulus, ...);
 *
 * DESCRIPTION: General function for sending a stimulus to CS and setting
 *              up the intention to await response(s) to that stimulus
 *              (all within a single process).
 *  
 *   The variable arguments are triples of the form:
 *    <context, retcode, response-string>, where "context" is the name of the
 *    state (a string) in which the response is enabled, "retcode" is the
 *    transition return code that triggers this response, and "response-string"
 *    is what to send back to TCM (either SUCCESS_RESPONSE or an arbitrary
 *    string, signifying failure).  The end of the list is indicated by the
 *    special symbol END_RESPONSES.
 *  
 *    e.g. tcmSendStimulus(node, fsm, "StartBB", "StartTraj", 1, "NoTrajectory",
 *                         "Completed", 0, SUCCESS_RESPONSE, END_RESPONSES);
 *
 * NOTES: Actually set up responses first, in case response comes immediately.
 *
 **************************************************************************/

TCM_Return_Type tcmSendStimulus (TCM_Task_Tree_Ref parentNode, CSFsm fsm,
				 String stimulus, ...)
{
  int numArgs, numResponses;
  TCM_CS_Response_Type responses[MAX_RESPONSES];
  va_list args;
  
  // Count how many responses there are.
  va_start(args, stimulus);
  for (numArgs = 0; va_arg(args, void *) != END_RESPONSES; numArgs++) {};
  va_end(args);

  numResponses = numArgs/3;
  if (numArgs == 0) {
    fprintf(stderr, "tcmSendStimulus: No responses for %s sent from %s\n",
	    stimulus, TCM_NodeName(parentNode));
    return TCM_Error;
  } else if (numArgs%3 != 0) {
    fprintf(stderr, 
	    "tcmSendStimulus: Uneven response triples for %s sent from %s\n",
	    stimulus, TCM_NodeName(parentNode));
    return TCM_Error;
  } else if (numResponses > MAX_RESPONSES) {
    fprintf(stderr, 
	    "tcmSendStimulus: Too many responses (%d) for %s sent from %s\n",
	    numResponses, stimulus, TCM_NodeName(parentNode));
  } else {
    va_start(args, stimulus);
    for (int i=0; i < numResponses; i++) {
      responses[i].set(va_arg(args, String), va_arg(args, int),
		       va_arg(args, String));
    }
    va_end(args);

    return tcmSendStimulus(parentNode, fsm, stimulus, numResponses, responses);
  }
}
