/*DBA-ALERT: Notes:
 *  Probable bugs with negative directions.  Particularly negative horizontal.
 *  Bug in scaling / destinations-part.
 *  Bug in scaling when decendents of children change.
 */

import java.io.PrintStream;
import netscape.application.*;
import netscape.util.*;


public class GraphConnection extends GraphComponent implements DragDestination
{
  public final static int   DEFAULT_HORIZONTAL_LENGTH  = 20; /* line length  */
  public final static int   DEFAULT_VERTICAL_LENGTH    = 20; /* line length  */
  public final static int   DEFAULT_LINE_SIZE          = 3;  /* line width   */
  public final static int   DEFAULT_SCALE_EXTRA_LENGTH = -15;/* shorten line */

  protected boolean         isDoubleLine; /*For parallel*/

	/** Are we auto-scaling? */
  protected boolean         isScaling;

	/** Amount to add to our length, auto-set when scaling. */
  protected int             scalingLength;

  protected boolean         canReceiveDrag;


  public GraphConnection ( )
  {
    super ( 1 ); /* We only have one child. */
    isDoubleLine   = false;
    isScaling      = false;
    scalingLength  = 0;
    canReceiveDrag = false;
    setParentDirection ( GraphComponent . UP );
  }

	/* Convenience initializer */
  public GraphConnection ( int             theParentDirection,
			   Color           theForegroundColor,
			   boolean         theCanRecieveDrag,
			   boolean         theIsScaling,
			   GraphComponent  theChild )
  {
    this();
    setParentDirection   ( theParentDirection );
    setForegroundColor   ( theForegroundColor );
    setCanReceiveDrag    ( theCanRecieveDrag  );
    setIsScaling         ( theIsScaling       );
    if ( theChild != null )
      addChild ( theChild );
  }

  public GraphComponent getSource ( ) { return getParent(); }
  public void           setSource ( GraphComponent theSource )
				      { setParent ( theSource ); }

  public GraphComponent getDestination ( ) { return getChild ( 0 ); }
  public void           setDestination ( GraphComponent theDestination )
  {
    removeChild ( 0 );
    addChild ( theDestination );
  }

  public boolean getIsHorizontal ( )
  {
    return ( GraphComponent . testDirection ( getParentDirection(),
					      GraphComponent.HORIZONTAL ) );
  }

  public boolean getIsVertical ( )
  {
    return ( GraphComponent . testDirection ( getParentDirection(),
					      GraphComponent.VERTICAL ) );
  }

  public boolean getIsPositiveDirection ( )
  {
    return ( GraphComponent . testDirection ( getParentDirection(),
					      GraphComponent.POSITIVE ) );
  }

  public boolean getIsNegativeDirection ( )
  {
    return ( GraphComponent . testDirection ( getParentDirection(),
					      GraphComponent.NEGATIVE ) );
  }


  public boolean getIsDoubleLine ( ) { return isDoubleLine; }
  public void    setIsDoubleLine ( boolean theIsDoubleLine )
  {
    isDoubleLine = theIsDoubleLine;
    realign();
  }

  public boolean getIsScaling ( ) { return isScaling;      }
  public void    setIsScaling ( ) { setIsScaling ( true ); }
  public void    setIsScaling ( boolean theIsScaling )
  {
    isScaling = theIsScaling;
    realign();
  }
  public int     getScalingLength ( ) 
  {
    return ( getIsScaling() ? scalingLength : 0 );
  }
  protected void setScalingLength ( int theScalingLength ) 
  {
    scalingLength = theScalingLength;
  }

  public boolean getCanReceiveDrag ( ) { return canReceiveDrag;      }
  public void    setCanReceiveDrag ( ) { setCanReceiveDrag ( true ); }
  public void    setCanReceiveDrag ( boolean theCanReceiveDrag )
				       { canReceiveDrag = theCanReceiveDrag; }



  public void realign ( )
  {
	/* Recompute our scaling... */
    checkScaling ();


	/* Figure out our width/height */
    if ( getIsHorizontal() == true )
    {
      setWidth  ( GraphConnection . DEFAULT_HORIZONTAL_LENGTH
		  + getScalingLength() );
      setHeight ( GraphConnection . DEFAULT_LINE_SIZE         );

      if ( getIsNegativeDirection() )
	setWidth ( getWidth() * -1 );
    }
    else /* Vertical line */
    {
      setWidth  ( GraphConnection . DEFAULT_LINE_SIZE       );
      setHeight ( GraphConnection . DEFAULT_VERTICAL_LENGTH
		  + getScalingLength() );

      if ( getIsNegativeDirection() )
	setHeight ( getHeight() * -1 );
    }


	/* Let the superclass do the work */
    super . realign();

  } /* public void realign() */



	/* This will be invoked after we have finished realigning, but
	 * before our children begin to get realigned.
	 */
  public void didRealign ( )
  {
	/* Hack to make width & height positive numbers */
    getComponentBounds();

	/* Let the superclass do the work */
    super . didRealign();
  }



	/* This will be invoked after both us and our children have finished
	 * realigning, but before our parent has been notified of the change.
	 */
  public void didRealignChildren ( GraphComponent theRealignedChildTree,
				   GraphComponent theOriginalChild      )
  {
    if ( checkScaling() == true )
    {
	/* Do everything again. */
      realign();
    }
    else
    {
	/* Let the superclass do its usual stuff */
      super . didRealignChildren ( theRealignedChildTree, theOriginalChild );
    }
  }


  public void parentDidRealignOtherChild ( GraphComponent theRealignedChild )
  {
    if ( checkScaling() == true )
    {
      realign();
    }
  }


  public boolean checkScaling ( )
  {
	/* Scale up the width or height up if we need to.. */
    if (   ( getIsScaling()   == true )
	&& ( getSource()      != null )
	&& ( getDestination() != null ) )
    {
      int  length = GraphConnection.DEFAULT_SCALE_EXTRA_LENGTH;
      Rect fullRect    = new Rect(),
	   limitedRect = new Rect();
      GraphComponent   source, destination;

	/* Swap source & destination for negative lines. */
      if ( getIsPositiveDirection() == true )
      {
	source      = getSource();
	destination = getDestination();
      }
      else /* We are going negative. */
      {
	source      = getDestination();
	destination = getSource();
      }


	/* Do the Scaling */
      if ( getIsHorizontal() == true )
      {
	  /* Scale based on our source-node's other children. */
	if (   ( source . getChildrenSubgraph     ( fullRect, this ) )
	    && ( source . computeChildrenSubgraph ( limitedRect,
						    GraphComponent.VERTICAL )
		) )
	{
	  length += fullRect.x + fullRect.width
	             - ( limitedRect.x + ( limitedRect.width / 2 ) );
	}

          /* Scale based on our destination-node's other children */
	if ( destination . computeChildrenSubgraph ( limitedRect,
						     GraphComponent.VERTICAL
						    ) )
	{
	  length += limitedRect.width / 2;
	}
      }

      else /* Vertical Line */
      {
	  /* Scale based on our source-node's other children. */
	if (   ( source . getChildrenSubgraph     ( fullRect, this ) )
	    && ( source . computeChildrenSubgraph ( limitedRect,
						    GraphComponent.HORIZONTAL )
		) )
	{
	  length += fullRect.y + fullRect.height
	             - ( limitedRect.y + ( limitedRect.height / 2 ) );
	}

          /* Scale based on our destination-node's other children */
	if ( destination . computeChildrenSubgraph ( limitedRect,
						     GraphComponent.HORIZONTAL
						    ) )
	{
	  length += limitedRect.height / 2;
	}
      }

	/* And save the new scaling length */
      if ( length != getScalingLength() )
      {
	setScalingLength ( length );
	return true;
      }

    } /* if ( we can/should scale ) */

    return false;
  }




    /** Paints this object and all of its children. */
  public void drawSelf ( Graphics theGraphics )
  {
	/* Idiocy check */
    if ( GraphComponent.testDirection ( getParentDirection(),
					GraphComponent.ALL   ) == false )
    {
      System.err.println ( "[GraphConnection:drawSelf]  Warning:  "
			   + "Illegal value for getParentDirection().  ("
			   + getParentDirection() + ")" );
    }


	/* Draw self */
    if ( getIsHorizontal() == true )
    {
      if ( getIsDoubleLine() == false )
      {
	theGraphics . drawLine ( getX(), midY(), maxX(), midY() );
      }
      else /* getIsDoubleLine() == true */
      {
	theGraphics . drawLine ( getX(), midY() - 1, maxX(), midY() - 1 );
	theGraphics . drawLine ( getX(), midY() + 1, maxX(), midY() + 1 );
      }
    }
    else /* Vertical line */
    {
      if ( getIsDoubleLine() == false )
      {
	theGraphics . drawLine ( midX(), getY(), midX(), maxY() );
      }
      else /* getIsDoubleLine() == true */
      {
	theGraphics . drawLine ( midX() - 1, getY(), midX() - 1, maxY() );
	theGraphics . drawLine ( midX() + 1, getY(), midX() + 1, maxY() );
      }
    }
  }




	/* Personal Drag & Drop extension interface */
  public DragDestination acceptsDrag ( DragSession theDragSession,
				       int theX, int theY )
  {
    GraphModule  ourGraphModule;

    if (   ( getCanReceiveDrag() == false )

	|| ( theDragSession      == null )

	|| ( ( (theDragSession . data()) instanceof GraphComponentSupplier )
	     == false )

	|| ( ( (GraphComponentSupplier) (theDragSession . data()) )
	     . supportsGraphType ( getGraphType() )
	     == false )
	)
    {
      return null;
    }

    if ( ((theDragSession . source()) instanceof GraphComponent) )
    {
      ourGraphModule = getOwningGraphModule();

      if ( ourGraphModule != null )
      {
	if ( ourGraphModule . getIsOffspringModuleOf (
			       (GraphComponent) (theDragSession . source()) ) )
	{
	  return null;
	}
      }
      else /* ourGraphModule == null */
      {
	if ( getIsAncestor ( (GraphComponent) (theDragSession . source()) ) )
	  return null;
      }
    }

    return this;
  }



	/* DragDestination interface */
  public boolean dragEntered ( DragSession theDragSession )
  {
    setSelected ( false );
    return true;
  }

  public boolean dragMoved ( DragSession theDragSession )
  {
    return true;
  }

  public void dragExited ( DragSession theDragSession )
  {
    setUnselected();
  }

  public boolean dragDropped ( DragSession theDragSession )
  {
    if ( acceptsDrag ( theDragSession, 0, 0 ) == null )
    {
      return false;
    }

    if ( GraphModule
	   . addChildAt ( this,
			  ((GraphComponentSupplier) (theDragSession . data()))
			    . getDataComponent(),
			  false, true )
	 == false )
    {
      return false;
    }

	/* Unselect us.. (We were selected during the drag/drop.) */
    setUnselected();


    return true;
  }




    /** For Debugging/Testing..  Its nice to see what things are... */
  public void print ( PrintStream theStream )
  {
    theStream . println ( "isHorizontalLine........: " + getIsHorizontal() );
    theStream . println ( "isVerticalLine..........: " + getIsVertical() );
    theStream . println ( "isPositiveDirection.....: "
			  + getIsPositiveDirection() );
    theStream . println ( "isNegativeDirection.....: "
			  + getIsNegativeDirection() );
    theStream . println ( "isDoubleLine............: " + getIsDoubleLine() );
    theStream . println ( "isScaling...............: " + getIsScaling() );
    theStream . println ( "scalinglength...........: " + getScalingLength() );
    super . print ( theStream );
  }
}
