
import netscape.util.*;

/*
 * GraphModule's act to manage a set of graph-subcomponents.
 * (Which in turn represent a single underlying Data-Component object.)
 *
 * firstComponent -- What this module attaches to a graph to become part
 *                   of that graph.
 * lastComponent  -- Where another module attches to this module to become
 *                   the next (unrelated) statement in that graph.
 * childComponent -- Where any contained children-modules are attached.
 * secondaryChildComponent -- Extraneous component connection location.
 *                            (Used for "else" in "if-else".)
 *
 * dataComponent -- The DataComponent this module represents.
 * moduleCanBeRemoved -- Disables module deletion method.
 * moduleIsRegisteredToObserve -- Insures registration integrity.
 * data -- Miscellaneous objects that we wish to keep track of.
 */

public class GraphModule extends Object implements DataComponentObserver
{
	/* Constants */
  final static int NONE_COMPONENT            = 0;
  final static int FIRST_COMPONENT           = 1;
  final static int LAST_COMPONENT            = 2;
  final static int CHILD_COMPONENT           = 4;
  final static int SECONDARY_CHILD_COMPONENT = 8;


	/* Class Methods */
  public static void findSelectedModules ( GraphComponent theGraphComponent,
					   Vector         theVector )
  {
    GraphModule owningGraphModule;

	/* Idiocy check */
    if ( theGraphComponent == null )
    {
      System.err.println ( "[GraphModule:findSelectedModules]  Warning:  "
			   + "theGraphComponent == null" );
      return;
    }

	/* Don't even bother finding the GraphModule for GraphRoot objects...*/
    if ( theGraphComponent instanceof GraphRoot )
      owningGraphModule = null;
    else
      owningGraphModule = theGraphComponent . getOwningGraphModule();


	/* If we don't have a GraphModule to work with... */
	/* Or we are continuing past this GraphModule...  */
    if (   ( owningGraphModule                       == null              )
	|| ( owningGraphModule . getLastComponent()  == theGraphComponent )
	|| ( owningGraphModule . getChildComponent() == theGraphComponent )
	|| ( owningGraphModule . getSecondaryChildComponent()
	                                             == theGraphComponent ) )
    {
      for ( int i=0;   i < theGraphComponent . getMaxChildren();   i++ )
      {
	if ( theGraphComponent . getChild ( i ) != null )
	{
	  findSelectedModules( theGraphComponent . getChild ( i ), theVector );
	}
      }
    }

    else  /* Otherwise, deal with this graph-module... */

    {
      if (   ( owningGraphModule . getIsSelected()         )
	  && ( owningGraphModule . getModuleCanBeRemoved() ) )
      {
	theVector . addElement ( owningGraphModule );
      }
      else /* Check the children if the parent was not selected/removable... */
      {
	if ( owningGraphModule . getChildComponent() != null )
	  findSelectedModules ( owningGraphModule . getChildComponent(),
				theVector );

	if ( owningGraphModule . getSecondaryChildComponent() != null )
	  findSelectedModules ( owningGraphModule.getSecondaryChildComponent(),
				theVector );
      }

	/* Check whatever comes after this owningGraphModule */
      if ( owningGraphModule . getLastComponent() != null )
	findSelectedModules ( owningGraphModule . getLastComponent(),
			      theVector );
    }

  }


  public static boolean addChildAt ( GraphComponent theGraphComponent,
				     DataComponent  theNewChildDataComponent,
				     boolean        theUseDelayedUpdate,
				     boolean        theIgnorePrevious    )
  {
    GraphModule    currentModule, ancestorModule;
    int            descendentType;

	/* Find out which module we currently belong to... */
    currentModule = theGraphComponent.getOwningGraphModule();

	/* Error check that */
    if ( currentModule == null )
    {
      System.err.println ( "[GraphModule:addChildAt]  Error:  "
			   + "theGraphComponent does not belong to a "
			   + "GraphModule...  Aborting..." );
      return false;
    }

	/* Find the GraphModule we decend from... */
    switch ( currentModule . getGraphComponentType ( theGraphComponent ) )
    {
      case GraphModule.CHILD_COMPONENT:
	ancestorModule = currentModule;
	descendentType = GraphModule.CHILD_COMPONENT;
	break;

      case GraphModule.SECONDARY_CHILD_COMPONENT:
	ancestorModule = currentModule;
	descendentType = GraphModule.SECONDARY_CHILD_COMPONENT;
	break;

      case GraphModule.LAST_COMPONENT:
	ancestorModule = currentModule . getAncestorGraphModule();
	descendentType
	  = currentModule . getLastAncestorGraphModuleDescendentType();
	break;

      default:
	System.err.println ( "[GraphModule:addChildAt]  Error:  "
		     + "theGraphComponent is not a TRAILING GraphComponent "
		     + "of its GraphModule.  Assuming LAST_COMPONENT." );
	ancestorModule = currentModule . getAncestorGraphModule();
	descendentType
	  = currentModule . getLastAncestorGraphModuleDescendentType();
	break;
    }

	/* Error check that... */
    if (   ( descendentType != GraphModule.CHILD_COMPONENT           )
	&& ( descendentType != GraphModule.SECONDARY_CHILD_COMPONENT ) )
    {
      System.err.println ( "[GraphModule:addChildAt]  Error:  "
			   + "UNKNOWN ChildType of Ancestor Module...  (" +
			   + descendentType + ")  Aborting..." );
      return false;
    }


	/* Add the new DataComponent to the parent */
	/* (Which will trigger the update) */
    return ancestorModule . addChild ( descendentType,
				       currentModule,
				       theNewChildDataComponent,
				       theUseDelayedUpdate,
				       theIgnorePrevious    );
  }





	/* Instance Variables */
  protected GraphComponent  firstComponent;
  protected GraphComponent  lastComponent;
  protected GraphComponent  childComponent;
  protected GraphComponent  secondaryChildComponent;
  protected DataComponent   dataComponent;
  protected boolean         moduleCanBeRemoved;
  protected boolean         moduleIsRegisteredToObserve;
  protected int             graphType;
  protected int             lastFoundAncestorGraphModuleParentType;

  GraphModule ()
  {
    firstComponent              = null;
    lastComponent               = null;
    childComponent              = null;
    secondaryChildComponent     = null;
    dataComponent               = null;
    moduleCanBeRemoved          = true;
    moduleIsRegisteredToObserve = false;
    graphType                   = GraphRoot.GUI_TREE_GRAPH;
    lastFoundAncestorGraphModuleParentType = GraphModule.NONE_COMPONENT;
  }

	/* Convenience initializer */
  GraphModule ( GraphComponent  theFirstComponent,
		GraphComponent  theLastComponent,
		GraphComponent  theChildComponent,
		GraphComponent  theSecondaryChildComponent,
		DataComponent   theDataComponent,
		boolean         theModuleCanBeRemoved,
		int             theGraphType )
  {
    this();
    setFirstComponent          ( theFirstComponent          );
    setLastComponent           ( theLastComponent           );
    setChildComponent          ( theChildComponent          );
    setSecondaryChildComponent ( theSecondaryChildComponent );
    setDataComponent           ( theDataComponent           );
    setModuleCanBeRemoved      ( theModuleCanBeRemoved      );
  }


  public GraphComponent getFirstComponent() { return firstComponent; }
  public GraphComponent getLastComponent()  { return lastComponent; }
  public GraphComponent getChildComponent() { return childComponent; }
  public GraphComponent getSecondaryChildComponent()
                                            { return secondaryChildComponent; }

  public GraphComponent getChild ( int theChildIndex )
  {
    switch ( theChildIndex )
    {
      case GraphModule.CHILD_COMPONENT:
	return getChildComponent();

      case GraphModule.SECONDARY_CHILD_COMPONENT:
	return getSecondaryChildComponent();

      default:
	System.err.println (
		    "[GraphModule:getChild]  Warning:  No child for index ("
		    + theChildIndex + ").  Using Primary Child." );
	return getChildComponent();
    }
  }


  public GraphComponent  getGraphComponent ( int theGraphComponent )
  {
	/* Find our component */
    switch ( theGraphComponent )
    {
      case GraphModule.FIRST_COMPONENT:
	return getFirstComponent();

      case GraphModule.LAST_COMPONENT:
	return getLastComponent();

      case GraphModule.CHILD_COMPONENT:
	return getChildComponent();

      case GraphModule.SECONDARY_CHILD_COMPONENT:
	return getSecondaryChildComponent();
    }

    System.err.println ( "[GraphModule:getGraphComponent]  Error:  "
			 + "Invalid value for theGraphComponent.  ("
			 + theGraphComponent + ")" );
    return null;
  }


  public int  getGraphComponentType ( GraphComponent theGraphComponent )
  {
    if ( theGraphComponent == null )
      return GraphModule.NONE_COMPONENT;

    if ( theGraphComponent == getFirstComponent() )
      return GraphModule.FIRST_COMPONENT;

    if ( theGraphComponent == getLastComponent() )
      return GraphModule.LAST_COMPONENT;

    if ( theGraphComponent == getChildComponent() )
      return GraphModule.CHILD_COMPONENT;

    if ( theGraphComponent == getSecondaryChildComponent() )
      return GraphModule.SECONDARY_CHILD_COMPONENT;

    return GraphModule.NONE_COMPONENT;
  }



  public boolean getIsLeadingComponent ( GraphComponent theGraphComponent )
  {
    return (   ( theGraphComponent != null                )
	    && ( theGraphComponent == getFirstComponent() ) );
  }

  public boolean getIsTrailingComponent ( GraphComponent theGraphComponent )
  {
    return (   (     theGraphComponent != null                           )
	    && (   ( theGraphComponent == getLastComponent()           )
		|| ( theGraphComponent == getChildComponent()          )
		|| ( theGraphComponent == getSecondaryChildComponent() ) ) );
  }

  public boolean getIsOffspringComponent ( GraphComponent theGraphComponent )
  {
    return (   (     theGraphComponent != null                           )
	    && (   ( theGraphComponent == getChildComponent()          )
		|| ( theGraphComponent == getSecondaryChildComponent() ) ) );
  }

	/** Is theGraphComponent's GraphModule our (possibly great) ancestor?*/
  public boolean getIsOffspringModuleOf ( GraphComponent theGraphComponent )
  {
    if ( theGraphComponent == null )
    {
      System.err.println ( "[GraphModule:getIsOffspringModuleOf]  Error:  "
			   + "theGraphComponent == null" );
      return false;
    }

    GraphModule theGraphModule = theGraphComponent . getOwningGraphModule();
    GraphModule ancestorGraphModule, oldAncestorGraphModule;

    if ( theGraphModule == null )
    {
      System.err.println ( "[GraphModule:getIsOffspringModuleOf]  Error:  "
			   + "theGraphModule == null" );
      return false;
    }

    ancestorGraphModule    = this;
    oldAncestorGraphModule = null;

    while (   ( ancestorGraphModule != null                   )
	   && ( ancestorGraphModule != theGraphModule         )
	   && ( ancestorGraphModule != oldAncestorGraphModule ) )
    {
      oldAncestorGraphModule = ancestorGraphModule;
      ancestorGraphModule    = ancestorGraphModule.getAncestorGraphModule();
    }


    if ( ancestorGraphModule != theGraphModule )
      return false;
    else
      return true;
  }



  public void setFirstComponent ( GraphComponent theGraphComponent )
  {
    if ( firstComponent != null )
      firstComponent . setOwningGraphModule ( null );
    firstComponent = theGraphComponent;
    if ( firstComponent != null )
      firstComponent . setOwningGraphModule ( this );
  }

  public void setLastComponent     ( GraphComponent theGraphComponent )
  {
    lastComponent = theGraphComponent;
  }

  public void setChildComponent    ( GraphComponent theGraphComponent )
  {
    childComponent = theGraphComponent;
  }

  public void setSecondaryChildComponent( GraphComponent theGraphComponent )
  {
    secondaryChildComponent = theGraphComponent;
  }


  public DataComponent getDataComponent() { return dataComponent; }
  public void          setDataComponent( DataComponent theDataComponent )
  {
    boolean   wasRegistered = getModuleIsRegisteredToObserve();

    if ( wasRegistered )
      unregisterToObserve();

    dataComponent = theDataComponent;

    if ( wasRegistered )
      registerToObserve();
  }


  public boolean getModuleCanBeRemoved ()  { return moduleCanBeRemoved; }
  public void    setModuleCanBeRemoved ( boolean theModuleCanBeRemoved )
		          { moduleCanBeRemoved = theModuleCanBeRemoved; }


  public boolean getModuleIsRegisteredToObserve()
					{ return moduleIsRegisteredToObserve; }
  protected void setModuleIsRegisteredToObserve ( boolean theIsRegistered )
			  { moduleIsRegisteredToObserve = theIsRegistered; }


  public int  getGraphType (                  ) { return graphType;         }
  public void setGraphType ( int theGraphType ) { graphType = theGraphType; }



	/** Returns the prior GraphModule before firstComponent. */
  public GraphModule  getPreviousGraphModule()
  {
    GraphModule  priorGraphModule;


    if ( getFirstComponent() == null )
    {
      System.err.println ( "[GraphModule:getPreviousGraphModule]  Error:  "
			   + "getFirstComponent() == null" );
      return null;
    }

    if ( getFirstComponent() . getParent() == null )
    {
      System.err.println ( "[GraphModule:getPreviousGraphModule]  Error:  "
			   + "getFirstComponent() . getParent() == null" );
      return null;
    }

	/* Just die on the GraphRoot... */
    if ( getFirstComponent() . getParent() instanceof GraphRoot )
    {
      return null;
    }


    priorGraphModule
      = getFirstComponent() . getParent() . getOwningGraphModule();


    if ( priorGraphModule == null )
    {
      System.err.println ( "[GraphModule:getPreviousGraphModule]  Error:  "
			   + "priorGraphModule == null." );
      return null;
    }

    if ( priorGraphModule == this )
    {
      System.err.println ( "[GraphModule:getPreviousGraphModule]  Error:  "
			   + "priorGraphModule == this." );
      return null;
    }

    return priorGraphModule;
  }

	/** Returns the nth-Parent node above us from which we are decended
	  * as a child and not as a last-component.
	  * In the case of error, it return the highest GraphModule that it
	  * found, which could be "this".
	  */
  public GraphModule getAncestorGraphModule()
  {
    GraphModule   currentGraphModule  = this;
    GraphModule   previousGraphModule;

	/* Reset the last-found component type */
    lastFoundAncestorGraphModuleParentType = GraphModule.NONE_COMPONENT;

    while ( true ) /* ENDLESS LOOP */
    {
	/* Idiocy tests... */
      if ( currentGraphModule . getFirstComponent() == null )
      {
	System.err.println ( "[GraphModule:getAncestorGraphModule]  Error:  "
		     + "currentGraphModule . getFirstComponent() == null" );
	return currentGraphModule;
      }

      if ( currentGraphModule . getFirstComponent() . getParent() == null )
      {
	System.err.println ( "[GraphModule:getAncestorGraphModule]  Error:  "
	  + "currentGraphModule . getFirstComponent() . getParent() == null" );
	return currentGraphModule;
      }

	/* Get the previous GraphModule */
      previousGraphModule = currentGraphModule . getPreviousGraphModule();

	/* If there was a problem... */
      if ( previousGraphModule == null )
      {
	return currentGraphModule;
      }

	/* If this is our (desired) ancestor... */
      if (     previousGraphModule . getChildComponent()
	   ==  currentGraphModule  . getFirstComponent() . getParent() )
      {
	lastFoundAncestorGraphModuleParentType
	  = GraphModule.CHILD_COMPONENT;
	return previousGraphModule;	
      }

      else if (     previousGraphModule . getSecondaryChildComponent()
	        ==  currentGraphModule  . getFirstComponent() . getParent() )
      {
	lastFoundAncestorGraphModuleParentType
	  = GraphModule.SECONDARY_CHILD_COMPONENT;
	return previousGraphModule;
      }

	/* Continue onwards... */
      currentGraphModule = previousGraphModule;
    }
  }

  public int getLastAncestorGraphModuleDescendentType()
  {
	/* Return the cached type... */
    return lastFoundAncestorGraphModuleParentType;
  }



	/** Count how many subsequent GraphModules come after lastComponent. */
  public int getCountNextGraphModules ( int  theGraphComponent )
  {
    GraphComponent graphComponent;
    int            count = 0;

	/* Idiocy check */
    if (   ( theGraphComponent != GraphModule.LAST_COMPONENT            )
	&& ( theGraphComponent != GraphModule.CHILD_COMPONENT           )
	&& ( theGraphComponent != GraphModule.SECONDARY_CHILD_COMPONENT ) )
    {
      System.err.println ( "[GraphModule:getCountNextGraphModules]  Error:  "
			   + "Invalid value for theGraphComponent.  ("
			   + theGraphComponent + ")" );
      return -1;
    }

    graphComponent = getGraphComponent ( theGraphComponent );

	/* Do we have a component to work off of? */
    if ( graphComponent == null )
    {
      System.err.println ( "[GraphModule:getCountNextGraphModules]  Error:  "
			   + "graphComponent == null" );
      return -1;
    }

    return graphComponent . getCountChildren();
  }

	/** Returns the nth subsequent GraphModule after lastComponent */
  public GraphModule  getNextGraphModule ( int  theNthGraphModule,
					   int  theGraphComponent )
  {
    GraphComponent childGraphComponent = null;
    GraphComponent graphComponent;
    GraphModule    nextGraphModule;

	/* Idiocy check */
    if (   ( theGraphComponent != GraphModule.LAST_COMPONENT            )
	&& ( theGraphComponent != GraphModule.CHILD_COMPONENT           )
	&& ( theGraphComponent != GraphModule.SECONDARY_CHILD_COMPONENT ) )
    {
      System.err.println ( "[GraphModule:getNextGraphModule]  Error:  "
			   + "Invalid value for theGraphComponent.  ("
			   + theGraphComponent + ")" );
    }

    graphComponent = getGraphComponent ( theGraphComponent );

	/* Do we have a component to work off of? */
    if ( graphComponent == null )
    {
      System.err.println ( "[GraphModule:getNextGraphModule]  Error:  "
			   + "graphComponent == null" );
      return null;
    }


	/* Find the Nth child... */
    for ( int i=0;  i < graphComponent . getMaxChildren();  i++ )
    {
      childGraphComponent = graphComponent . getChild ( i );

      if ( childGraphComponent != null )
      {
	if ( theNthGraphModule <= 0 )
	  break;
	else
	  theNthGraphModule --;
      }
    }

	/* Not enough children... */
    if (   ( theNthGraphModule   >  0    )
	|| ( childGraphComponent == null ) )
    {
      return null;
    }


    nextGraphModule = childGraphComponent . getOwningGraphModule();


    if ( nextGraphModule == null )
    {
      System.err.println ( "[GraphModule:getNextGraphModule]  Error:  "
			   + "nextGraphModule == null." );
      return null;
    }

    if ( nextGraphModule == this )
    {
      System.err.println ( "[GraphModule:getNextGraphModule]  Error:  "
			   + "nextGraphModule == this." );
      return null;
    }

    return nextGraphModule;
  }


	/* Select everything except trailing components.
	 * Don't select past lastComponent.
	 */
  public void recursivelySetSelected ( GraphComponent  theGraphComponent )
  {
    GraphComponent  childGraphComponent;
    GraphModule     childGraphModule;

	/* Open up any GraphMeets... */
    if ( theGraphComponent instanceof GraphMeet )
      ((GraphMeet) theGraphComponent) . setIsOpen();

    for ( int i=0;   i < theGraphComponent . getMaxChildren();   i++ )
    {
      childGraphComponent = theGraphComponent . getChild ( i );

      if (   ( childGraphComponent != null               )
	  && ( childGraphComponent != getLastComponent() ) )
      {
	childGraphModule = childGraphComponent . getOwningGraphModule();

	if (   ( childGraphModule == null )
	    || ( childGraphModule
		  . getIsTrailingComponent ( childGraphComponent ) == false ) )
	{
	  childGraphComponent . setSelected ( false, false );
	}
	recursivelySetSelected ( childGraphComponent );
      }
    }
  }


	/* Unselect everything except lastComponent.
	 * (Including other trailing components.)
	 * Don't select past lastComponent.
	 */
  public void recursivelySetUnselected ( GraphComponent  theGraphComponent,
					 GraphComponent  theBeneathComponent )
  {
	/* We can start unselecting if we have reached theBeneathComponent */
    if ( theGraphComponent == theBeneathComponent )
      theBeneathComponent = null;

    GraphComponent  childGraphComponent;

    for ( int i=0;   i < theGraphComponent . getMaxChildren();   i++ )
    {
      childGraphComponent = theGraphComponent . getChild ( i );

      if (   ( childGraphComponent != null               )
	  && ( childGraphComponent != getLastComponent() ) )
      {
	if (   ( theBeneathComponent == null                )
	    || ( theBeneathComponent == childGraphComponent ) )
	  childGraphComponent . setUnselected ( false, false );

	if ( childGraphComponent . getMaxChildren() > 0 )
	  recursivelySetUnselected ( childGraphComponent,
				     theBeneathComponent );
      }
    }
  }


  public boolean getIsSelected ( )
  {
    return (   ( getFirstComponent() != null           )
	    && ( getFirstComponent() . getIsSelected() ) );
  }


  public void setUnselected ( )  { setUnselected ( null ); }

  public void setUnselected ( GraphComponent  theBeneathComponent )
  {
    if ( getIsSelected() )
    {
	/* If we are selected,
	 * check to see if we are part of a larger selected grouping...
	 */
      GraphModule  ancestorGraphModule = getAncestorGraphModule();

      if (   ( ancestorGraphModule != null           )
	  && ( ancestorGraphModule != this           )
	  && ( ancestorGraphModule . getIsSelected() ) )
      {
	  /* If so, unselect that group */
	ancestorGraphModule . setUnselected ( theBeneathComponent );
      }
      else  /* Otherwise, unselect ourselves... */
      {
	getFirstComponent() . setUnselected ( false, false );
	recursivelySetUnselected ( getFirstComponent(), null );
	getFirstComponent() . redisplay ( null );

      }
    }
	
    else  /* Otherwise, unselect us anyway to catch offspring unselects... */
    {
      if ( getFirstComponent() != null )
      {
	recursivelySetUnselected ( getFirstComponent(), theBeneathComponent );
      }
      else
      {
	System.err.println ( "[GraphModule:setUnselected]  Error:  "
			     + "getFirstComponent() == null." );
      }
    }
  }


  public void toggleSelected ( boolean         theIsMultipleSelection,
			       GraphComponent  theSourceGraphComponent )
  {
    if ( getFirstComponent() == null )
    {
      System.err.println ( "[GraphModule:toggleSelected]  Error:  "
			   + "getFirstComponent() == null" );
      return;
    }

    if ( getIsTrailingComponent ( theSourceGraphComponent ) )
    {
      if (    theIsMultipleSelection 
	   && theSourceGraphComponent . getIsSelected() )
      {
	theSourceGraphComponent . setUnselected();
	return;
      }
      else
      {
	theSourceGraphComponent
	  . setSelected ( theIsMultipleSelection == false );
	return;
      }
    }

    if ( theIsMultipleSelection && getIsSelected() )
    {
      setUnselected();
    }
    else /* Otherwise let getFirstComponent().setSelected() do the work */
    {
      getFirstComponent() . setSelected ( theIsMultipleSelection == false,
					  false );
      recursivelySetSelected ( getFirstComponent() );
      getFirstComponent() . redisplay ( null );
    }
  }


	/** Adds new data to the underlying DataComponent */
  public boolean addChild ( int             theChildLocation,
			    GraphModule     theAddAfterThisChildModule,
			    DataComponent   theNewDataComponent,
			    boolean         theUseDelayedUpdate,
			    boolean         theIgnorePrevious    )
  {
    DataComponent  addChildAfterDataComponent;
    boolean        returnValue;

	/* Deal with case where this is our first child */
    if ( theAddAfterThisChildModule == this )
      addChildAfterDataComponent = null;
    else
      addChildAfterDataComponent
	= theAddAfterThisChildModule . getDataComponent();
      

    if (   ( theChildLocation != GraphModule.CHILD_COMPONENT           )
	&& ( theChildLocation != GraphModule.SECONDARY_CHILD_COMPONENT ) )
    {
      System.err.println ( "[GraphModule:addChild]  Warning:  "
			   + "Unknown value for theChildLocation ("
			   + theChildLocation
			   + ").  Assuming Primary Child-Branch." );
    }

    if ( theChildLocation != GraphModule.SECONDARY_CHILD_COMPONENT )
    {
      returnValue
	= getDataComponent() . addPrimaryChild ( theNewDataComponent,
						 addChildAfterDataComponent );
    }
    else /* Ie:  theChildLocation == GraphModule.SECONDARY_CHILD_COMPONENT */
    {
      returnValue
	= getDataComponent() . addSecondaryChild( theNewDataComponent,
						  addChildAfterDataComponent );
    }

    if ( returnValue )
    {
	/* Have DataComponent Update itself only if we succeeded... */
    if ( theUseDelayedUpdate )
      getDataComponent() . delayedUpdate ( theIgnorePrevious );
    else
      getDataComponent() . update();
    }
    else
    {
      System.err.println ( "[GraphModule:addChild]  Warning:  "
			   + "Add of child failed." );
    }

    return returnValue;
  }



	/* Removes our underlying DataComponent */
  public boolean removeSelf ( boolean theUseDelayedUpdate,
			      boolean theIgnorePrevious   )
  {
    DataComponent  parentDataComponent;

    if ( getDataComponent() == null )
    {
      System.err.println ( "[GraphModule:removeSelf]  Warning:  "
			   + "No DataComponent available to remove." );
      return false;
    }

    parentDataComponent = getDataComponent() . getParent();

    if ( parentDataComponent == null )
    {
      System.err.println ( "[GraphModule:removeSelf]  Warning:  "
			 + "No DataComponent PARENT available to remove us." );
      return false;
    }

    if ( parentDataComponent . removeChild ( getDataComponent() ) == false )
    {
      System.err.println ( "[GraphModule:removeSelf]  Warning:  "
			   + "DataComponent (parent) removeChild() failed." );
      return false;
    }

	/* Stop observing our DataComponent... */
    unregisterToObserve();

	/* Have the parent update itself to update everyone else... */
    if ( theUseDelayedUpdate )
      parentDataComponent . delayedUpdate ( theIgnorePrevious );
    else
      parentDataComponent . update();

    return true;
  }



	/* Graph-Update method */
  public void removeModuleFromGraph()
  {
	/* Idiocy checks */
    if ( getModuleCanBeRemoved() == false )
    {
      System.err.println ( "[GraphModule:removeModuleFromGraph]  Error:  "
			   + "Unable to remove un-removable module..." );
      return;
    }


    int              i;
    GraphComponent   child,
		     parent = getFirstComponent() . getParent();
                     

    parent . removeChild ( getFirstComponent() );
    for ( i=0;  i < getLastComponent() . getMaxChildren();  i++ )
    {
      child = getLastComponent() . getChild ( i );
      if ( child != null )
      {
	if ( getLastComponent() . removeChild ( i ) == false )
	{
	  System.err.println ( "[GraphModule:removeModuleFromGraph]  Error:  "
	     + "Unable to remove child (" + i + ") from LastComponent.  "
	     + "Child deleted as well..." );
	}
	else
	{
	  parent . addChild ( child );
	}
      } /* if ( child != null ) */
    } /* FOR ( 0 <= i < getLastComponent() . getChildren() ) */

	/* Lets stop watching stuff... */
    unregisterToObserve();
  }


	/* Removes this GraphModule and all Last-element trailing
	 * GraphModules from the Graph...
	 * (Requires that all the lastComponent's be GraphConnections.)
	 */
  public void removeModuleSubtreeFromGraph ( Vector theGraphModulesVector )
  {
    GraphModule     graphModule = this;
    GraphModule     newGraphModule;
    GraphComponent  graphComponent;
    final int       originalVectorSize = theGraphModulesVector . count();
    int             i, nth;


	/* Start with this graphModule */
    theGraphModulesVector . addElement ( this );

	/* Find all the GraphModules. */
    for ( i=originalVectorSize;  i < theGraphModulesVector . count();  i++ )
    {
      graphModule = (GraphModule) (theGraphModulesVector . elementAt ( i ) );

      for ( nth = 0;
	    nth < graphModule . getCountNextGraphModules (
						  GraphModule.LAST_COMPONENT );
	    nth ++ )
      {
	newGraphModule
	  = graphModule . getNextGraphModule ( nth,
					       GraphModule.LAST_COMPONENT );
	if ( newGraphModule != null )
	  theGraphModulesVector . addElement ( newGraphModule );
      }
    }

	/* Remove the GraphModules from the graph */
    for ( i  = theGraphModulesVector . count() - 1;
	  i >= originalVectorSize;
	  i -- )
    {
      ( (GraphModule) ( theGraphModulesVector . elementAt ( i ) ) )
	. removeModuleFromGraph();
    }

  }



	/* Graph-Update method */
  public boolean insertModuleAt ( GraphComponent theGraphComponent )
  {
    int             firstComponentDirection;
    GraphComponent  childToMove = null;

	/* Direction we are adding firstComponent to it's new parent */
    firstComponentDirection = getFirstComponent() . getParentDirection();

	/* Pre-existing child of said parent that needs moved. */
    childToMove = theGraphComponent . findChild ( firstComponentDirection );


	/* If we do NOT need to move a child */
    if ( childToMove == null )
    {
      if ( theGraphComponent . addChild ( getFirstComponent() ) )
      {
	registerToObserve();
	return true;
      }
      else
	return false;
    }
    else
    {
	/* Do we NOT have an opening? */	
      if ( getLastComponent() . findChildIndex ( firstComponentDirection )
	   != GraphComponent.INVALID )
      {
	System.err.println ( "[GraphModule:insertModuleAt]  Error:  "
	     + "Both target and last GraphComponent's have a child in the "
	     + "same direction.  (" + firstComponentDirection
	     + ")  Failing insertion..." );
	return false;
      }
      else /* We should/might have an opening */
      {
	    /* Remove the child */
	theGraphComponent . removeChild ( childToMove );

	   /* And try to add it back to ourself */
	if ( getLastComponent() . addChild ( childToMove ) == false )
	{
	  System.err.println ( "[GraphModule:insertModuleAt]  Error:  "
	     + "Unable to add child to lastComponent.  Failing insertion..." );
	  if ( theGraphComponent . addChild ( childToMove ) == false )
	  {
	    System.err.println ( "[GraphModule:insertModuleAt]  Error:  "
	       + "Unable to restore removed child.  "
	       + "Double failure during insertion." );
	  }
	  return false;
	}

	    /* Try to add us in there */
	if ( theGraphComponent . addChild ( getFirstComponent() ) == false )
	{
	  System.err.println ( "[GraphModule:insertModuleAt]  Error:  "
	     + "Unable to add firstComponent.  Failing insertion..." );

	  if ( getLastComponent() . removeChild ( childToMove ) == false )
	  {
	    System.err.println ( "[GraphModule:insertModuleAt]  Error:  "
	       + "Unable to remove child prior to child-restoration.  "
	       + "DOUBLE Failure during insertion." );
	  }

	  if ( theGraphComponent . addChild ( childToMove ) == false )
	  {
	    System.err.println ( "[GraphModule:insertModuleAt]  Error:  "
	       + "Unable to restore removed child.  "
	       + "TRIPLE Failure during insertion." );
	  }
	  return false;
	}
	else  /* It worked.  We are done. */
	{
	  registerToObserve();
	  return true;
	}
      } /* IF ( we don't have an opening ) ...  else  .... */
    } /* IF ( childToMove == null ) ... else ... */
  }


  public GraphModule findGraphModule ( DataComponent theForDataComponent,
				       int           theSearchDirections )
  {
    GraphComponent   graphComponent;
    GraphModule      relativeGraphModule, foundGraphModule;
    int              i, nth, checkDirection;


	/* Trivial Case */
    if ( theForDataComponent == getDataComponent() )
      return this;


	/* First Component case */
    if ( ( theSearchDirections & GraphModule.FIRST_COMPONENT )   != 0 )
    {
      relativeGraphModule = getPreviousGraphModule();

		/* Go the whole way up, and then start going back down... */
      if ( relativeGraphModule != null )
      {
	return relativeGraphModule
	        . findGraphModule ( theForDataComponent, theSearchDirections );
      }
    }

	/* Stop searching upwards...   No endless loops */
    if ( ( theSearchDirections & GraphModule.FIRST_COMPONENT ) != 0 )
      theSearchDirections ^= GraphModule.FIRST_COMPONENT;



    for ( i=0;  i<3;  i++ )
    {
	/* Configure Search Direction */
      switch ( i )
      {
	case 0:
	  checkDirection  = GraphModule.LAST_COMPONENT;
	  graphComponent  = getLastComponent();
	  break;

	case 1:
	  checkDirection  = GraphModule.CHILD_COMPONENT;
	  graphComponent  = getChildComponent();
	  break;

	case 2:
	default:
	  checkDirection  = GraphModule.SECONDARY_CHILD_COMPONENT;
	  graphComponent  = getSecondaryChildComponent();
	  break;
      }
      

	/* Search in given direction */
      if (   ( ( theSearchDirections & checkDirection )  != 0    )
	  && (   graphComponent                          != null ) )
      {
	for ( nth=0;
	      nth < getCountNextGraphModules ( checkDirection );
	      nth++ )
	{
	  relativeGraphModule = getNextGraphModule ( nth, checkDirection );

	  if ( relativeGraphModule != null )
	  {
	    foundGraphModule
	      = relativeGraphModule 
	        . findGraphModule ( theForDataComponent, theSearchDirections );

	    if ( foundGraphModule != null )
	      return foundGraphModule;
	  }
	} /* For ( 0 <= nth < getCountNextGraphModules ( checkDirection ) ) */

      } /* IF ( theSearchDirections & checkDirection && graphComponent!=null */
    } /* for ( i=0;  i<3;  i++ ) */

	/* If we get here, we failed... */
    return null;
  }



	/* DataComponentObserver Interface */
  public boolean update ( DataComponent  theObservedComponent,
			  DataComponent  theModifiedComponent,
			  Object         theStoredData        )
  {
    if ( theObservedComponent != getDataComponent() )
    {
      System.err.println ( "[GraphModule:update]  Warning:  "
	       + "theObservedComponent != getDataComponent().   Aborting..." );
      return false;
    }

	/* No sense in implement a different GraphModule class for each */
	/* possible DataComponent we could be build for...              */
    return theObservedComponent . updateGraphModule ( theModifiedComponent,
						      theStoredData,
						      this );
  }


  public boolean registerToObserve ()
  {
    if ( getDataComponent() . addObserver ( this ) )
    {
      setModuleIsRegisteredToObserve ( true );
      return true;
    }
    else
      return false;
  }

  public boolean unregisterToObserve ()
  {
    if ( getDataComponent() . removeObserver ( this ) )
    {
      setModuleIsRegisteredToObserve ( false );
      return true;
    }
    else
      return false;
  }

  public boolean disableUpdates ( boolean thePrintWarning )
  {
    if (   ( getFirstComponent()               != null )
	&& ( getFirstComponent() . getParent() != null ) )
    {
      getFirstComponent() . getParent() . disableRealignPropagationUp();
      return true;
    }
    else
    {
      if ( thePrintWarning )
      {
	System.err.println (
	  "[GraphModule:disableUpdates]  Warning:  Unable to disableUpdates.  "
	  + "No parent of firstComponent available." );
      }
      return false;
    }
  }

  public boolean reenableUpdates ( boolean thePrintWarning )
  {
    if (   ( getFirstComponent()               != null )
	&& ( getFirstComponent() . getParent() != null ) )
    {
      getFirstComponent() . getParent() . reenableRealignPropagationUp();
      getFirstComponent() . realign();
      return true;
    }
    else
    {
      if ( thePrintWarning )
      {
	System.err.println (
	  "[GraphModule:reenableUpdates]  Warning:  Unable to reenableUpdates."
	  + "  No parent of firstComponent available." );
      }
      return false;
    }
  }

}

