// Font.java
// By Ned Etcode
// Copyright 1995, 1996, 1997 Netscape Communications Corp.  All rights reserved.

package netscape.application;

import java.net.URL;
import java.awt.image.FilteredImageSource;
import java.awt.image.CropImageFilter;
import java.awt.image.ImageFilter;

import netscape.util.*;

/** Object subclass representing a Font.
  */


public class Font implements Codable {
    FontMetrics                 _metrics;
    String                      _name;
    URL                         _url;
    Bitmap                      _glyphsImage;
    Vector                      _glyphVector, _widthsVector;
    Hashtable                   _description;
    java.awt.Font               _awtFont;
    int                         _type, _widthsArrayBase, _widthsArray[];

    final static int            INVALID = 0, AWT = 1, DOWNLOADED = 2;

    /** Plain font style. */
    public final static int     PLAIN = java.awt.Font.PLAIN;
    /** Bold font style. */
    public final static int     BOLD = java.awt.Font.BOLD;
    /** Italic font style. */
    public final static int     ITALIC = java.awt.Font.ITALIC;

    final static String         FAMILY = "Family", STYLE = "Style",
                                SIZE = "Size", WIDTHS = "Widths";
    final static String         DESCRIPTION = "Description",
                                GLYPHS = "glyphs.gif";

    final static String         NAME_KEY = "name";
    final static String         STYLE_KEY = "style";
    final static String         SIZE_KEY = "size";
    private static Class fontClass;

    private static Class fontClass() {
        if (fontClass == null) {
            fontClass = new Font().getClass();
        }

        return fontClass;
    }

    /** Constructs an empty Font. This constructor is only useful when
      * decoding.
      */
    public Font() {
        super();
        _type = INVALID;
    }

    /** Constructs a Font with the specified name, style and size.
      */
    public Font(String name, int style, int size) {
        this();

        _awtFont = new java.awt.Font(name, style, size);
        _name = name;
        if (_awtFont != null) {
            _type = AWT;
        }
    }

    /** Returns the "default" font, currently defined to be Helvetica 12.
      */
    public static Font defaultFont() {
        return fontNamed("Helvetica", Font.PLAIN, 12);
    }


final static int[]  courierPlainFontSizes    = { 1,2,  1,4,  2,5,  2,5,  3,6,  4,8,  4,9,  5,10,  6,12,  6,13,  7,14,  7,15,  8,16,  9,17,  9,18,  10,20,  11,21,  11,22,  12,24,  12,24,  13,25,  14,27,  14,28,  15,29,  16,30,  16,32,  17,33,  17,33,  18,35,  19,36,  19,37,  20,39,  21,40,  21,41,  22,42  };

final static int[]  courierBoldFontSizes     = { 1,2,  1,4,  2,5,  2,6,  3,6,  4,8,  4,9,  5,10,  6,12,  6,13,  7,14,  7,16,  8,16,  9,17,  9,18,  10,20,  11,21,  11,22,  12,24,  12,25,  13,25,  14,27,  14,28,  15,29,  16,30,  16,32,  17,33,  17,34,  18,36,  19,36,  19,37,  20,39,  21,40,  21,41,  22,42  };

final static int[]  courierItalicFontSizes   = { 1,1,  1,2,  2,3,  3,5,  3,6,  4,7,  4,8,  5,10,  6,10,  6,11,  6,11,  7,14,  9,15,  9,16,  10,16,  10,16,  11,16,  11,19,  11,18,  12,21,  12,19,  13,20,  14,21,  15,25,  16,25,  16,26,  17,27,  18,31,  18,29,  19,30,  19,31,  20,30,  21,30,  21,31,  22,32  };

final static int[]  helveticaPlainFontSizes  = { 1,2,  2,3,  3,4,  5,6,  6,7,  7,8,  8,8,  9,11,  10,11,  11,13,  11,13,  12,15,  14,17,  15,17,  16,17,  17,19,  18,19,  18,22,  18,21,  22,25,  20,23,  21,23,  22,24,  25,28,  26,28,  27,29,  28,30,  30,33,  30,33,  31,34,  32,35,  34,36,  35,36,  36,37,  37,38  };

final static int[]  helveticaBoldFontSizes   = { 1,2,  2,3,  3,4,  5,6,  6,7,  7,8,  8,8,  9,11,  10,11,  11,13,  11,13,  13,15,  15,17,  15,17,  16,17,  16,19,  17,19,  18,22,  18,22,  22,25,  20,24,  21,24,  22,26,  24,28,  25,28,  26,29,  27,30,  30,33,  29,33,  30,34,  31,35,  33,36,  34,36,  35,37,  36,38  };

final static int[]  helveticaItalicFontSizes = { 1,2,  2,3,  3,4,  5,6,  6,7,  7,8,  8,8,  9,11,  10,11,  11,13,  11,13,  12,15,  15,17,  15,17,  16,17,  17,19,  18,19,  18,22,  18,22,  22,25,  20,24,  21,24,  22,26,  23,28,  24,28,  25,29,  26,30,  30,33,  28,33,  29,34,  30,35,  34,36,  35,36,  36,37,  37,38  };

final static int      maxFontHackSize  = 36;
final static boolean  ENABLE_DEBUGGING = true;


    /** Retrieves the Font for the specified font name, style, and size, and
      * places it in the Application's font cache, allowing future requests
      * for the font to be handled efficiently.
      */
    public static synchronized Font fontNamed(String fontName, int style,
                                              int size) {
        Font font;
        String uniqueName;
        Application app;

        if (fontName == null || size == 0) {
            return null;
        }

        uniqueName = fontName + "." + style + "." + size;

        app = Application.application();
        font = (Font)app.fontByName.get(uniqueName);

        if (font != null)
            return font;

        font = new Font(fontName, style, size);
        if (!font.isValid())
            return null;

/*
System.err.println ( "Invoking: fontNamed(\"" + fontName + "\"," + style + "," + size + ")" );
*/

  if (   (    fontName . equals ( "Helvetica" )
	   || fontName . equals ( "Courier"   ) )
      && ( size >  1                            )
      && ( size <  maxFontHackSize              ) )
  {
    int  targetWidth, targetHeight;

    if ( fontName . equals ( "Courier" ) )
    {
      switch ( style )
      {
	case Font.BOLD:
	  targetWidth  = courierBoldFontSizes   [ ( (size - 1) * 2 )     ];
	  targetHeight = courierBoldFontSizes   [ ( (size - 1) * 2 ) + 1 ];
	  break;

	case Font.ITALIC:
	  targetWidth  = courierItalicFontSizes [ ( (size - 1) * 2 )     ];
	  targetHeight = courierItalicFontSizes [ ( (size - 1) * 2 ) + 1 ];
	  break;

	case Font.PLAIN:   
	default:
	  targetWidth  = courierPlainFontSizes  [ ( (size - 1) * 2 )     ];
	  targetHeight = courierPlainFontSizes  [ ( (size - 1) * 2 ) + 1 ];
	  break;
      }
    }
    else
    {
      switch ( style )
      {
	case Font.BOLD:
	  targetWidth  = helveticaBoldFontSizes   [ ( (size - 1) * 2 )     ];
	  targetHeight = helveticaBoldFontSizes   [ ( (size - 1) * 2 ) + 1 ];
	  break;

	case Font.ITALIC:
	  targetWidth  = helveticaItalicFontSizes [ ( (size - 1) * 2 )     ];
	  targetHeight = helveticaItalicFontSizes [ ( (size - 1) * 2 ) + 1 ];
	  break;

	case Font.PLAIN:   
	default:
	  targetWidth  = helveticaPlainFontSizes  [ ( (size - 1) * 2 )     ];
	  targetHeight = helveticaPlainFontSizes  [ ( (size - 1) * 2 ) + 1 ];
	  break;
      }
    }

    FontMetrics fontMetrics = new FontMetrics ( font );
    Font        oldFont = font;
    Font        newFont = font;
    int         newSize = size;

    while (   (   ( fontMetrics.maxAdvance() > targetWidth  )
	       || ( fontMetrics.height()     > targetHeight ) )
	   && ( newSize > 1                                   ) )
    {
      newSize --;
      oldFont = newFont;
      newFont = new Font ( fontName, style, newSize );
      if ( !newFont.isValid() )
      {
	if ( Font.ENABLE_DEBUGGING )
	{
	  System.err.println( "Font:  Unable to acquire substitution font [\""
			      + fontName +"\"," + style + "," + newSize 
			      + "] for old font [ ..., " + size + "]" );
	}
	  /* Restore old values */
	newSize ++;
	newFont = oldFont;
	break;
      }
      fontMetrics = new FontMetrics ( newFont );

    } /* while ( Font is too big & newSize can be shrunk ) */

    if ( ( newSize != size )  &&  Font.ENABLE_DEBUGGING )
    {
      System.err.println ( "Font:  Rescaling font [\""
			   + fontName +"\"," + style + "," + size
			   + "] to size " + newSize );
    }

	/* Replace the current font with the new, scaled, Font... */
    font = newFont;

  } /* if ( we can rescale this font ) */



/*
FontMetrics fontMetrics = new FontMetrics ( font );
System.err.println ( "          metrics:  "
		     + "\n           height......: " + fontMetrics.height()
		     + "\n           leading.....: " + fontMetrics.leading() 
		     + "\n           ascent......: " + fontMetrics.ascent()
		     + "\n           descent.....: " + fontMetrics.descent()
		     + "\n           height......: " + fontMetrics.height()
		     + "\n           charHeight..: " + fontMetrics.charHeight()
		     + "\n           maxAscent...: " + fontMetrics.maxAscent()
		     + "\n           maxDescent..: " + fontMetrics.maxDescent()
		     + "\n           maxAdvance..: " + fontMetrics.maxAdvance()
		     );
*/

        app.fontByName.put(uniqueName, font);

        return font;
    }

    /** Returns the Font identified by the String <b>aString</b>. Font names
      * normally take the form "FontName:style:size". This is a convience for
      * fontNamed() as the string is parsed apart and
      * then fontNamed(name, style, size) is called.
      */
    public static Font fontNamed(String aString) {
        String[]        fontDef;
        String          fontName, fontStyleString;
        char            firstChar;
        int             fontSize, fontStyle;

        fontDef = stringsForString(aString);
        if (fontDef == null || fontDef.length == 0) {
            return null;
        }

        fontName = fontDef[0];
        if (fontName.equals("Default")) {
            return defaultFont();
        } else if (fontName.length() == 0) {
            return null;
        }

        /* if just a name, it's a downloadable font */
        if (fontDef.length == 1) {
            URL url;
            Application app = Application.application();

            url = app._appResources.urlForFontNamed(fontName);
            return getFontFromURL(url, fontName);
        }

        fontStyleString = fontDef[1];
        firstChar = fontStyleString.charAt(0);
        if (firstChar == 'P' || firstChar == 'p') {
            fontStyle = Font.PLAIN;
        } else if (firstChar == 'B' || firstChar == 'b') {
            fontStyle = Font.BOLD;
        } else if (firstChar == 'I' || firstChar == 'i') {
            fontStyle = Font.ITALIC;
        } else {
            try {
                fontStyle = Integer.parseInt(fontStyleString);
            } catch (NumberFormatException e) {
                fontStyle = Font.PLAIN;
            }
        }

        fontSize = parseInt(fontDef[2]);
        if (fontSize < 0) {
            fontSize = 0;
        }

        return fontNamed(fontName, fontStyle, fontSize);
    }

    private static int parseInt(String s) {
        try {
            return Integer.parseInt(s);
        } catch (NumberFormatException e) {
            return 0;
        }
    }

    private static String[] stringsForString(String aString) {
        Vector          stringVector;
        String[]        strings;
        String          string0, string1, string2;
        StringBuffer    tmpBuffer;
        int             i, count;

        if (aString == null) {
            return null;
        }

        if (aString.indexOf(':') == -1) {
            strings = new String[1];
            strings[0] = aString;
            return strings;
        }

        count = aString.length();
        tmpBuffer = new StringBuffer();
        stringVector = new Vector();
        for (i = 0; i < count; i++) {
            if (aString.charAt(i) == ':') {
                stringVector.addElement(tmpBuffer.toString());
                tmpBuffer = new StringBuffer();
            } else {
                tmpBuffer.append(aString.charAt(i));
            }
        }
        stringVector.addElement(tmpBuffer.toString());

        count = stringVector.count();
        strings = new String[count];
        for (i = 0; i < count; i++) {
            strings[i] = (String)stringVector.elementAt(i);
        }
        return strings;
    }

    // This is used in decoding.  Take a look at this again to see if it
    // causes problems by having multiple fonts with the same name.  ALERT!

    synchronized void nameFont(String fontName, Font font) {
        Application.application().fontByName.put(fontName, font);
    }

    static synchronized Font getFontFromURL(URL absoluteURL, String fontName) {
        Font font;
        Application app = Application.application();

        if (absoluteURL == null) {
            return null;
        }

        if (fontName != null) {
            font = (Font)app.fontByName.get(fontName);
            if (font != null)
                return font;
        }

        font = new Font(absoluteURL, fontName);
        if (!font.isValid())
            return null;

        app.fontByName.put(fontName, font);

        return font;
    }

    /** Convenience constructor for instantiating a downloadable Font with
      * the specified name and URL locating the Font's resources on the web
      * server.
      */
    Font(java.net.URL fontURL, String name) {
        this();

        URL                     configURL;
        java.io.InputStream     inputStream;
        Deserializer            deserializer;

        _url = fontURL;
        _name = name;
        _type = DOWNLOADED;

        _glyphVector = new Vector();
        _widthsVector = new Vector();

      /* description file */
        try {
            configURL = new URL(fontURL, _name + ".font/" + DESCRIPTION);
        } catch (Exception e) {
            System.err.println(
                "Font.init() - Trouble creating font description URL " +
                        fontURL + _name + ".font/" + DESCRIPTION + " : " + e);
            _type = INVALID;
            return;
        }

        try {
            inputStream = configURL.openStream();
            deserializer = new Deserializer(inputStream);
            _description = (Hashtable)deserializer.readObject();

            _loadWidths();
        } catch (Exception e) {
            System.err.println(
                "Font.init() - Trouble retrieving font description URL " +
                                                    configURL);
            e.printStackTrace(System.err);
            _type = INVALID;
            return;
        }

      /* glyphs image */
        _loadGlyphs(fontURL);
    }

    void _loadWidths() {
        Object    widthsInfo[];
        int       i, count, index;

        if (_description == null) {
            return;
        }

        widthsInfo = (Object[])_description.get(WIDTHS);

        if (widthsInfo == null) {
            System.err.println(
                "Font._loadWidths() - No widths information for " + _name);
            return;
        }

        _widthsArrayBase = parseInt((String)widthsInfo[0]);

        count = widthsInfo.length - 1;
        _widthsArray = new int[_widthsArrayBase + count];
        for (i = 0; i < _widthsArrayBase; i++) {
            _widthsArray[i] = 0;
        }
        for (i = 1, index = _widthsArrayBase; i < count; i++, index++) {
            _widthsArray[index] = parseInt((String)widthsInfo[i]);
        }

        if (_widthsArray[' '] == 0) {
            _widthsArray[' '] = 5;
        }
    }

    java.awt.Image croppedImage(int x, int y, int width, int height) {
        java.awt.Image  croppedImage;

        croppedImage = AWTCompatibility.awtApplet().createImage(
            new FilteredImageSource(_glyphsImage.awtImage.getSource(),
            new CropImageFilter(x, y, width, height)));

        return croppedImage;
    }

    void _loadGlyphs(java.net.URL fontURL) {
        java.net.URL                            glyphURL;
        Image                                glyphsImage, newImage;
        java.awt.image.ImageFilter              imageChopper;
        FilteredImageSource                     filteredImageSource;
        java.awt.Image                          awtImage;
        int                                     i, x, c;

      /* description file */
        try {
            glyphURL = new URL(fontURL, _name + ".font/" + GLYPHS);
        } catch (Exception e) {
            System.err.println(
                "Font.init() - Trouble creating font glyph URL " +
                        fontURL + _name + ".font/" + GLYPHS + " : " + e);
            _type = INVALID;
            return;
        }

        _glyphsImage = Bitmap.bitmapFromURL(glyphURL);
        _glyphsImage.loadData();

        if (_glyphsImage == null || !_glyphsImage.isValid()) {
            System.err.println(
                "Font._loadGlyphs() - Trouble loading glyphs for " + _name);
            return;
        }

        _glyphsImage.loadData();

        for (i = _widthsArrayBase, x = 0; i < _widthsArray.length; i++) {
            awtImage = croppedImage(x, 0, _widthsArray[i],
                                     _glyphsImage.height());
            _glyphVector.addElement(
                                AWTCompatibility.bitmapForAWTImage(awtImage));
            newImage = (Image)_glyphVector.lastElement();

            x += _widthsArray[i];
        }
    }

    /** Returns <b>true</b> if the Font is valid.  A Font will not
      * be valid if there were problems creating its java.awt.Font (for local
      * Fonts), or there was a problem downloading the Font's resources
      * (for downloadable Fonts).
      */
    boolean isValid() {
        return (_type != INVALID);
    }

    /** Returns <b>true</b> if the Font was downloaded from the Applet's
      * web server.
      */
    boolean wasDownloaded() {
        return (_type == DOWNLOADED);
    }

    /** Returns the Font's font metrics. */
    public FontMetrics fontMetrics() {
        if (_metrics == null) {
            _metrics = new FontMetrics(this);
        }

        return _metrics;
    }

    /** Returns the Font's family. */
    public String family() {
        if (_type == INVALID) {
            return "";
        }

        if (_awtFont != null) {
            return _awtFont.getFamily();
        } else {
            return (String)_description.get(FAMILY);
        }
    }

    /** Returns the Font's name.  For local fonts this is the font name,
      * such as "Helvetica".
      */
    public String name() {
        if (_type == INVALID) {
            return "";
        }

        return _name;
    }

    /** Returns the integer representation of the Font's style. */
    public int style() {
        if (_type == INVALID) {
            return -1;
        }

        if (_awtFont != null) {
            return _awtFont.getStyle();
        } else {
            return parseInt((String)_description.get(STYLE));
        }
    }

    /** Returns the Font's point size. */
    public int size() {
        if (_type == INVALID) {
            return -1;
        }

        if (_awtFont != null) {
            return _awtFont.getSize();
        } else {
            return parseInt((String)_description.get(SIZE));
        }
    }

    /** Returns <b>true</b> if the Font's style is "plain." */
    public boolean isPlain() {
        return ((style() == PLAIN) ? true : false);
    }

    /** Returns <b>true</b> if the Font's style is "bold." */
    public boolean isBold() {
        return (((style() & BOLD) > 0) ? true : false );
    }

    /** Returns <b>true</b> if the Font's style is "italic."
      */
    public boolean isItalic() {
        return (((style() & ITALIC) > 0) ? true : false );
    }

    /** Returns the Font's "glyph vector," its collection of images that
      * comprise the font.  This method is valid only with a downloaded font.
      */
    Vector glyphVector() {
        return _glyphVector;
    }

    String _stringValueFromDescription(String keyName) {
        if (keyName == null || _description == null) {
            return "";
        }

        return (String)_description.get(keyName);
    }

    int _intValueFromDescription(String keyName) {
        if (keyName == null || _description == null) {
            return 0;
        }

        return parseInt((String)_description.get(keyName));
    }

    /** Returns the Font's string representation. The string takes the form
      * "Fontname.style.size" and is suitable for use in Font.fontNamed().
      */
    public String toString() {
        String          styleString;

        if (_type == INVALID || wasDownloaded()) {
            return _name;
        }

        if (isBold()) {
            if (isItalic()) {
                styleString = "BoldItalic";
            } else {
                styleString = "Bold";
            }
        } else if (isItalic()) {
            styleString = "Italic";
        } else {
            styleString = "Plain";
        }

        return family() + ":" + styleString + ":" + size();
    }

    /** Describes the Font class' information.
      * @see Codable#describeClassInfo
      */
    public void describeClassInfo(ClassInfo info) {
        info.addClass("netscape.application.Font", 1);
        info.addField(NAME_KEY, STRING_TYPE);
        info.addField(STYLE_KEY, INT_TYPE);
        info.addField(SIZE_KEY, INT_TYPE);
    }

    /** Encodes the Font instance by name.
      * @see Codable#encode
      */
    public void encode(Encoder encoder) throws CodingException {
        encoder.encodeString(NAME_KEY, name());
        encoder.encodeInt(STYLE_KEY, style());
        encoder.encodeInt(SIZE_KEY, size());
    }

    /** Decodes the Font instance by name.
      * @see Codable#decode
      */
    public void decode(Decoder decoder) throws CodingException {
        Font    font;
        String  name;
        int style, size;

        name = decoder.decodeString(NAME_KEY);
        style = decoder.decodeInt(STYLE_KEY);
        size = decoder.decodeInt(SIZE_KEY);

        font = fontNamed(name, style, size);

        // Only replace the object if we are not a subclass of Font.

        if (getClass() == fontClass()) {
            decoder.replaceObject(font);
        } else {
            _name = font._name;
            _type = font._type;
            _awtFont = font._awtFont;
            nameFont(_name, this);
        }
    }

    /** Finishes the Font's decoding.
      * @see Codable#finishDecoding
      */
    public void finishDecoding() throws CodingException {
    }


  public static void main ( String args[] )
  {
    Font         font;
    FontMetrics  fontMetrics;
    int          i, styleIndex, style, size;

    for ( i=0;  i < args.length;  i++ )
    {
      System.out.println ( "Font:  \"" + args[i] + "\"" );

      for ( styleIndex=0;  styleIndex < 3;  styleIndex ++ )
      {
	switch ( styleIndex )
	{
	  case 1:  style = Font.BOLD;    break;
	  case 2:  style = Font.ITALIC;  break;
	  case 0:
	  default: style = Font.PLAIN;   break;
	}

	System.out.print ( "{ " );

	for ( size = 1;  size < maxFontHackSize;  size ++ )
	{
	  font = new Font ( args[i], style, size );
	  if ( font == null )
	  {
	    System.err.println( "Invalid Font:  (\"" + args[i] + "\","
				+ style + "," + size + ")" );
	    return;
	  }

	  fontMetrics = new FontMetrics ( font );
	  if ( fontMetrics == null )
	  {
	    System.err.println( "Invalid FontMetrics:  (\"" + args[i] + "\","
				+ style + "," + size + ")" );
	    return;
	  }

	  System.out.print ( fontMetrics.maxAdvance() + ","
			     + fontMetrics.height() );
	  if ( (size+1) < maxFontHackSize )
	    System.out.print ( ",  " );
	  else
	    System.out.print ( "  }\n" );

	} /* for ( size = 1;  size < maxFontHackSize;  size ++ ) */
      } /* for ( styleIndex=0;  styleIndex < 3;  styleIndex ++ ) */
    } /* for ( i=0;  i < args.length;  i++ ) */

    System.out.println ( "Completed Successfully..." );

    return;

  } /* public static int main ( String args[] ) */
}
